# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from .equine import Equine, EquineOutput
from .equine_gp import EquineGP
from .equine_protonet import EquineProtonet, CovType

from .utils import (
    brier_score,
    brier_skill_score,
    expected_calibration_error,
    generate_support,
)

__version__ = "0.1.1rc1"

__all__ = [
    "Equine",
    "EquineOutput",
    "EquineGP",
    "EquineProtonet",
    "CovType",
    "brier_score",
    "brier_skill_score",
    "expected_calibration_error",
    "generate_support",
]



# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import equine as eq
import torch
import os
from hypothesis import given, settings

from conftest import BasicEmbeddingModel, random_dataset


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=10)
def test_equine_gp_train_from_scratch(random_dataset) -> None:
    dataset, num_classes, _ = random_dataset
    X, _ = dataset.tensors
    embedding_model = BasicEmbeddingModel(X.shape[1], num_classes)

    model = eq.EquineGP(embedding_model, num_classes, num_classes)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    model.train_model(dataset, loss_fn, optimizer, 10)

    model.predict(X[1:10])  # Contracts should fire asserts on errors


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=10)
def test_equine_gp_train_from_scratch_with_temperature(random_dataset) -> None:
    dataset, num_classes, _ = random_dataset
    X, _ = dataset.tensors
    embedding_model = BasicEmbeddingModel(X.shape[1], num_classes)

    model = eq.EquineGP(embedding_model, num_classes, num_classes, use_temperature=True)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    _, cal_loader = model.train_model(dataset, loss_fn, optimizer, 10)

    model.predict(X[1:10])  # Contracts should fire asserts on errors

    model.calibrate_temperature(cal_loader, 1, 0.01)


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=2)
def test_equine_gp_save_load(random_dataset) -> None:
    dataset, num_classes, _ = random_dataset
    X, _ = dataset.tensors
    embedding_model = BasicEmbeddingModel(X.shape[1], num_classes)

    model = eq.EquineGP(embedding_model, num_classes, num_classes)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    model.train_model(dataset, loss_fn, optimizer, 10)

    old_output = model.predict(X[1:10])
    tmp_filename = "tmp_eq_gp.pt"
    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)
    model.save(tmp_filename)
    new_model = eq.EquineGP.load(tmp_filename)
    new_output = new_model.predict(X[1:10])
    assert (
        torch.nn.functional.mse_loss(old_output.classes, new_output.classes) <= 1e-7
    ), "Predictions changed on reload"
    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=2)
def test_equine_gp_save_load_with_temperature(random_dataset) -> None:
    dataset, num_classes, _ = random_dataset
    X, _ = dataset.tensors
    embedding_model = BasicEmbeddingModel(X.shape[1], num_classes)

    model = eq.EquineGP(embedding_model, num_classes, num_classes, use_temperature=True)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    model.train_model(dataset, loss_fn, optimizer, 10)

    old_output = model.predict(X[1:10])
    tmp_filename = "tmp_eq_gp.pt"
    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)
    model.save(tmp_filename)
    new_model = eq.EquineGP.load(tmp_filename)
    new_output = new_model.predict(X[1:10])
    assert (
        torch.nn.functional.mse_loss(old_output.classes, new_output.classes) <= 1e-7
    ), "Predictions changed on reload"
    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup



# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
from __future__ import annotations
from typing import Callable, Any
import io
from enum import Enum
from collections import OrderedDict
from datetime import datetime

import icontract
import torch
from torch.utils.data import TensorDataset
from typeguard import typechecked
from sklearn.model_selection import train_test_split
from scipy.stats import gaussian_kde
from tqdm import tqdm

from .equine import Equine, EquineOutput
from .utils import generate_support, generate_episode, generate_train_summary


#####################################
class CovType(Enum):
    UNIT = "unit"
    DIAGONAL = "diag"
    FULL = "full"


PRED_COV_TYPE = CovType.DIAGONAL
OOD_COV_TYPE = CovType.DIAGONAL
DEFAULT_EPSILON = 1e-5
COV_REG_TYPE = "epsilon"


###############################################
@typechecked
class _Protonet(torch.nn.Module):
    """
    Private class that implements a prototypical neural network for use in EQUINE.
    """

    def __init__(
        self,
        embedding_model: torch.nn.Module,
        emb_out_dim: int,
        cov_type: CovType,
        cov_reg_type: str,
        epsilon: float,
    ) -> None:
        """Protonet class constructor
        :param embedding_model: The pytorch embedding model to generate logits with
        :param emb_out_dim: Dimention size of given embedding model's output
        :param cov_type: Type of covariance to use when computing distances [unit, diag, full]
        :param cov_reg_type: type of regularization to use when generating the covariance matrix [epsilon, shared]
        :param epsilon: Epsilon value to use for covariance regularization
        """
        super().__init__()
        self.embedding_model = embedding_model
        self.cov_type = cov_type
        self.cov_reg_type = cov_reg_type
        self.epsilon = epsilon
        self.emb_out_dim = emb_out_dim

        self.support = None
        # self.support_embeddings = None
        self.model_head = self.create_model_head(emb_out_dim)

    def create_model_head(self, emb_out_dim: int):
        """Method for adding a pytorch layer on top of given embedding model
        :param emb_out_dim: Dimention size of embedding model output
        :return torch.nn.Linear: Returns the created pytorch model layer
        """
        return torch.nn.Linear(emb_out_dim, emb_out_dim)

    def compute_embeddings(self, X: torch.Tensor) -> torch.Tensor:
        """Method for calculating model embeddings using both the given embedding model and the added model head.
        :param X: Input tensor to compute embeddings on
        :return torch.Tensor: Returns fully computed embedding tensors for given X tensor
        """
        model_embeddings = self.embedding_model(X)
        head_embeddings = self.model_head(model_embeddings)
        return head_embeddings

    @icontract.require(lambda self: self.support_embeddings is not None)
    def compute_prototypes(self) -> torch.Tensor:
        """Method for computing class prototypes based on given support examples
        :return torch.Tensor: Tensors of prototypes for each of the given classes in the support
        """
        # Compute prototype for each class
        proto_list = []
        for label in self.support_embeddings:  # look at doing functorch
            class_prototype = torch.mean(self.support_embeddings[label], dim=0)  # type: ignore
            proto_list.append(class_prototype)

        prototypes = torch.stack(proto_list)

        return prototypes

    @icontract.require(lambda self: len(self.support_embeddings) > 0)
    def compute_covariance(self, cov_type: CovType) -> torch.Tensor:
        """Method for generating the regularized support example covariance matrix used for calculating distances
        :param cov_type: Type of covariance to use [unit, diag, full]
        :return torch.Tensor: Tensor containing the generated regularized covariance matrix
        """
        class_cov_dict = OrderedDict().fromkeys(self.support_embeddings.keys())
        for label in self.support_embeddings.keys():
            class_covariance = self.compute_covariance_by_type(
                cov_type, self.support_embeddings[label]
            )
            class_cov_dict[label] = class_covariance

        reg_covariance_dict = self.regularize_covariance(class_cov_dict)
        reg_covariance = torch.stack(list(reg_covariance_dict.values()))

        return reg_covariance  # TODO try putting everything on GPU with .to() and see if faster

    def compute_covariance_by_type(
        self, cov_type: CovType, embedding: torch.Tensor
    ) -> torch.Tensor:
        """Method to select appropriate covariance matrix type based on cov_type
        :param cov_type: Type of covariance to use [unit, diag, full]
        :param embedding: embedding tensor to use when generating the covariance matrix
        :return torch.Tensor: Tensor containing the requested covariance matrix
        """
        if cov_type == CovType.FULL:
            class_covariance = torch.cov(embedding.T)
        elif cov_type == CovType.DIAGONAL:
            class_covariance = torch.var(embedding, dim=0)
        elif cov_type == CovType.UNIT:
            class_covariance = torch.ones(self.emb_out_dim)
        else:
            raise ValueError

        return class_covariance

    def regularize_covariance(
        self, class_cov_dict: dict[float, torch.Tensor]
    ) -> dict[float, torch.Tensor]:
        """Method to add regularization to each class covariance matrix based on the selected regularization type
        :param class_cov_dict: A dictionary containing each class and the corresponding covariance matrix
        :return dict: Dictonary containing the regularized class covariance matrices
        """
        if self.cov_type == CovType.FULL:
            regularization = torch.diag(self.epsilon * torch.ones(self.emb_out_dim))
        elif self.cov_type == CovType.DIAGONAL:
            regularization = self.epsilon * torch.ones(self.emb_out_dim)
        elif self.cov_type == CovType.UNIT:
            regularization = torch.zeros(self.emb_out_dim)
        else:
            raise ValueError("Unknown Covariance Type")

        if self.cov_reg_type == "shared":
            if self.cov_type != CovType.FULL and self.cov_type != CovType.DIAGONAL:
                raise ValueError(
                    "Covariance type FULL and DIAGONAL are incompatible with shared regularization"
                )

            shared_covariance = self.compute_shared_covariance(class_cov_dict)

            for label in self.support_embeddings:
                num_class_support = self.support_embeddings[label].shape[0]
                lamb = num_class_support / (num_class_support + 1)

                class_cov_dict[label] = (
                    lamb * class_cov_dict[label]
                    + (1 - lamb) * shared_covariance
                    + regularization
                )

        elif self.cov_reg_type == "epsilon":
            for label in self.support_embeddings:
                class_cov_dict[label] = class_cov_dict[label] + regularization

        return class_cov_dict

    def compute_shared_covariance(
        self, class_cov_dict: dict[float, torch.Tensor]
    ) -> torch.Tensor:
        """Method to calculate a shared covariance matrix
        :param class_cov_dict: A dictionary containing each class and the corresponding covariance matrix
        :return torch.Tensor: Tensor containing the shared covariance matrix
        """
        total_support = sum([x.shape[0] for x in class_cov_dict.values()])

        if self.cov_type == CovType.FULL:
            shared_covariance = torch.zeros((self.emb_out_dim, self.emb_out_dim))
        elif self.cov_type == CovType.DIAGONAL:
            shared_covariance = torch.zeros(self.emb_out_dim)
        else:
            raise ValueError(
                "Shared covariance can only be used with FULL or DIAGONAL (not UNIT) covariance types"
            )

        for label in class_cov_dict:
            num_class_support = class_cov_dict[label].shape[0]
            shared_covariance = (
                shared_covariance + (num_class_support - 1) * class_cov_dict[label]
            )  # undo N-1 div from cov

        shared_covariance = shared_covariance / (
            total_support - 1
        )  # redo N-1 div for shared cov

        return shared_covariance

    @typechecked
    @icontract.require(lambda X_embed, mu: X_embed.shape[-1] == mu.shape[-1])
    @icontract.ensure(lambda result: torch.all(result >= 0))
    def compute_distance(
        self, X_embed: torch.Tensor, mu: torch.Tensor, cov: torch.Tensor
    ) -> torch.Tensor:
        """Method to compute the distances to class prototypes for the given embeddings
        :param X_embed: The embeddings to compute the distances on
        :param mu: The class prototypes
        :param cov: The support covariance matrix
        :return torch.Tensor: The calculated distances from each of the class prototypes for the given embeddings
        """
        _queries = torch.unsqueeze(X_embed, 1)  # examples x 1 x dimension
        diff = torch.sub(mu, _queries) ** 2  # examples x classes x dimension

        if len(cov.shape) == 2:  # (diagonal covariance)
            # examples x classes x dimension
            dist = torch.nan_to_num(torch.div(diff, cov))
            dist = torch.sum(dist, dim=2)  # examples x classes
            dist = dist.squeeze(dim=1)
            dist = torch.sqrt(dist + self.epsilon)  # examples x classes
        else:  # len(cov.shape) == 3: (full covariance)
            diff = diff.permute(1, 2, 0)  # classes x dimension x examples
            sol = torch.linalg.lstsq(cov, diff, rcond=10 ** (-4))
            sol = sol.solution  # classes x dimension x examples
            dist = torch.sum(diff * sol, dim=1)  # classes x examples
            dist = torch.sqrt(dist.permute(1, 0) + self.epsilon)  # examples x classes
            dist = dist.squeeze(dim=1)

        return dist

    @typechecked
    def compute_classes(self, distances: torch.Tensor) -> torch.Tensor:
        """Method to compute predicted classes from distances via a softmax function
        :param distances: The distances of embeddings to class prototypes
        :return torch.Tensor: Tensor of class predictions
        """
        softmax = torch.nn.functional.softmax(torch.neg(distances), dim=-1)
        return softmax

    @typechecked
    def forward(self, X: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """_Protonet forward function, generates class probability predictions
        :param X: Input tensor for generating predictions
        :return tuple[torch.Tensor, torch.Tensor]: Tuple containing class probability predictions, and class distances from prototypes
        """
        if self.support is None or self.support_embeddings is None:
            raise ValueError(
                "No support examples found. Protonet Model requires model support to \
                    be set with the 'update_support()' method before calling forward."
            )

        X_embed = self.compute_embeddings(X)
        if X_embed.shape == torch.Size([self.emb_out_dim]):
            X_embed = X_embed.unsqueeze(dim=0)  # handle single examples
        distances = self.compute_distance(X_embed, self.prototypes, self.covariance)
        classes = self.compute_classes(distances)

        return classes, distances

    def update_support(self, support: OrderedDict) -> None:
        """Method to update the support examples, and all the calculations that rely on them
        :param support: Ordered dict containing class labels and their associated support examples
        """
        self.support = support  # TODO torch.nn.ParameterDict(support)

        support_embs = OrderedDict().fromkeys(support.keys())
        for label in support:
            support_embs[label] = self.compute_embeddings(support[label])

        self.support_embeddings = (
            support_embs  # TODO torch.nn.ParameterDict(support_embs)
        )

        self.prototypes = self.compute_prototypes()

        if self.training is False:
            self.compute_global_moments()
            self.covariance = self.compute_covariance(cov_type=PRED_COV_TYPE)
        else:
            self.covariance = self.compute_covariance(cov_type=self.cov_type)

    @typechecked
    @icontract.require(lambda self: self.support_embeddings is not None)
    def compute_global_moments(self) -> None:
        """Method to calculate the global moments of the support embeddings for use in OOD score generation"""
        embeddings = torch.cat(list(self.support_embeddings.values()))
        self.global_covariance = torch.unsqueeze(
            self.compute_covariance_by_type(OOD_COV_TYPE, embeddings), dim=0
        )
        self.global_mean = torch.mean(embeddings, dim=0)


###############################################
@typechecked
class EquineProtonet(Equine):
    """
    An example of an EQUINE model that utilizes protonets and relative mahalanobis distances
    to generate OOD and model confidence scores. This wraps any pytorch embedding neural network
    and provides the `forward`, `predict`, `save`, and `load` methods required by Equine.
    """

    def __init__(
        self,
        embedding_model,
        emb_out_dim: int,
        cov_type: CovType = CovType.UNIT,
        relative_mahal: bool = True,
        use_temperature: bool = False,
        init_temperature: float = 1.0,
    ) -> None:
        """EquineProtonet constructor
        :param embedding_model: Neural Network feature embedding model
        :param emb_out_dim: The number of output features from the embedding model
        :param cov_type: The type of covariance to use when training the protonet [UNIT, DIAG, FULL]
        :param relative_mahal: Use relative mahalanobis distance for OOD calculations. If false, uses standard mahalanobis distance instead
        :param use_temperature: whether to use temperature scaling after training
        param init_temperature: what to use as the initial temperature (1.0 has no effect)
        """
        super().__init__(embedding_model)
        self.cov_type = cov_type
        self.cov_reg_type = COV_REG_TYPE
        self.relative_mahal = relative_mahal
        self.emb_out_dim = emb_out_dim
        self.epsilon = DEFAULT_EPSILON
        self.outlier_score_kde = None
        self.model_summary = None
        self.use_temperature = use_temperature
        self.init_temperature = init_temperature
        self.register_buffer(
            "temperature", torch.Tensor(self.init_temperature * torch.ones(1))
        )

        self.model = _Protonet(
            embedding_model,
            self.emb_out_dim,
            self.cov_type,
            self.cov_reg_type,
            self.epsilon,
        )

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """EquineProtonet forward function, generates logits for classification
        :param X: Input tensor for generating predictions
        :return torch.Tensor: Output class predictions
        """
        preds, _ = self.model(X)
        return preds

    @icontract.require(lambda calib_frac: calib_frac > 0 and calib_frac < 1)
    def train_model(
        self,
        dataset: TensorDataset,
        num_episodes: int,
        calib_frac: float = 0.2,
        support_size: int = 25,
        way: int = 3,
        episode_size: int = 100,
        loss_fn: Callable = torch.nn.functional.cross_entropy,
        opt_class: Callable = torch.optim.Adam,
        num_calibration_epochs: int = 2,
        calibration_lr: float = 0.01,
    ) -> tuple[dict[str, Any], torch.Tensor, torch.Tensor]:
        """Train or fine-tune an EquineProtonet model
        :param dataset: Input pytorch TensorDataset of training data for model
        :param num_episodes: The desired number of episodes to use for training
        :param calib_frac: Fraction of given training data to reserve for model calibration
        :param support_size: Number of support examples to generate for each class
        :param way: Number of classes to train on per episode
        :param episode_size: Number of examples to use per episode
        :param loss_fn: A pytorch loss function, eg., torch.nn.CrossEntropyLoss()
        :param opt_class: A pytorch optimizer, e.g., torch.optim.Adam
        :param num_calibration_epochs: The desired number of epochs to use for temperature scaling,
        :param calibration_lr: learning rate for temperature scaling
        :return: A tuple containing the model summary, the held out calibration data, and the calibration labels
        """
        self.train()

        if self.use_temperature:
            self.temperature = torch.Tensor(
                self.init_temperature * torch.ones(1)
            ).type_as(self.temperature)

        X, Y = dataset[:]

        train_x, calib_x, train_y, calib_y = train_test_split(
            X, Y, test_size=calib_frac, stratify=Y
        )  # TODO: Replace sklearn with torch call
        optimizer = opt_class(self.parameters())

        for i in tqdm(range(num_episodes)):
            optimizer.zero_grad()

            support, episode_x, episode_y = generate_episode(
                train_x, train_y, support_size, way, episode_size
            )
            self.model.update_support(support)

            _, dists = self.model(episode_x)
            loss_value = loss_fn(torch.neg(dists), episode_y)
            loss_value.backward()
            optimizer.step()

        self.eval()
        full_support = generate_support(
            train_x,
            train_y,
            support_size,
            selected_labels=torch.unique(train_y).tolist(),
        )

        self.model.update_support(
            full_support
        )  # update support with final selected examples

        X_embed = self.model.compute_embeddings(calib_x)
        pred_probs, dists = self.model(calib_x)
        ood_dists = self._compute_ood_dist(X_embed, pred_probs, dists)
        self._fit_outlier_scores(ood_dists, calib_y)

        if self.use_temperature:
            self.calibrate_temperature(
                calib_x, calib_y, num_calibration_epochs, calibration_lr
            )

        date_trained = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        self.train_summary = generate_train_summary(self, train_y, date_trained)
        return self.train_summary, calib_x, calib_y

    def calibrate_temperature(
        self,
        calib_x: torch.Tensor,
        calib_y: torch.Tensor,
        num_calibration_epochs: int = 1,
        calibration_lr: float = 0.01,
    ) -> None:
        """
        Fine-tune the temperature after training.  Note this function is also run at the conclusion of train_model
        :param calib_x: training data to be used for temperature calibration
        :param calib_y: labels corresponding to calib_x
        :param num_calibration_epochs: number of epochs to tune temperature
        :param calibration_lr: learning rate for temperature optimization
        :return:    None
        """
        self.temperature.requires_grad = True
        optimizer = torch.optim.Adam([self.temperature], lr=calibration_lr)
        for t in range(num_calibration_epochs):
            optimizer.zero_grad()
            with torch.no_grad():
                pred_probs, dists = self.model(calib_x)
            dists = dists / self.temperature
            loss = torch.nn.functional.cross_entropy(
                torch.neg(dists), calib_y.to(torch.long)
            )
            loss.backward()
            optimizer.step()
        self.temperature.requires_grad = False

    @icontract.ensure(lambda self: self.model.support_embeddings is not None)
    def _fit_outlier_scores(
        self, ood_dists: torch.Tensor, calib_y: torch.Tensor
    ) -> None:
        """Private function to fit outlier scores with a kernel density estimate (KDE)
        :param ood_dists: Tensor of computed OOD distances
        :param calib_y: Tensor of class labels for ood_dists examples
        """
        self.outlier_score_kde = OrderedDict.fromkeys(
            self.model.support_embeddings.keys()
        )

        for label in self.outlier_score_kde:
            class_ood_dists = ood_dists[calib_y == int(label)].detach().numpy()
            class_kde = gaussian_kde(class_ood_dists)  # TODO convert to torch func
            self.outlier_score_kde[label] = class_kde

    def _compute_outlier_scores(self, ood_dists, predictions) -> torch.Tensor:
        """Private function to compute OOD scores using the calculated kernel density estimate (KDE)
        :param ood_dists: Tensor of computed OOD distances
        :param predictions: Tensor of model protonet predictions
        :return torch.Tensor: Tensor of OOD scores for the given examples
        """
        ood_scores = torch.zeros_like(ood_dists)
        for i in range(len(predictions)):
            # Use KDE and RMD corresponding to the predicted class
            predicted_class = int(torch.argmax(predictions[i, :]))
            p_value = self.outlier_score_kde[int(predicted_class)].integrate_box_1d(
                ood_dists[i].detach().numpy(), torch.inf
            )
            ood_scores[i] = 1.0 - p_value

        return ood_scores

    @icontract.ensure(lambda result: len(result) > 0)
    def _compute_ood_dist(
        self,
        X_embeddings: torch.Tensor,
        predictions: torch.Tensor,
        distances: torch.Tensor,
    ) -> torch.Tensor:
        """Private function to compute OOD distances using a distance function
        :param X_embeddings: Tensor of example embeddings
        :param predictions: Tensor of model protonet predictions for the given embeddings
        :param distances: Tensor of calculated protonet distances for the given embeddings
        :return torch.Tensor: Tensor of OOD distances for the given embeddings
        """
        preds = torch.argmax(predictions, dim=1)
        preds = preds.unsqueeze(dim=-1)
        # Calculate (Relative) Mahalanobis Distance:
        if self.relative_mahal:
            null_distance = self.model.compute_distance(
                X_embeddings, self.model.global_mean, self.model.global_covariance
            )
            null_distance = null_distance.unsqueeze(dim=-1)
            ood_dist = distances.gather(1, preds) - null_distance
        else:
            ood_dist = distances.gather(1, preds)

        ood_dist = torch.reshape(ood_dist, (-1,))
        return ood_dist

    def predict(self, X: torch.Tensor) -> EquineOutput:
        """Predict function for EquineProtonet, inherited and implemented from Equine
        :param X: Input tensor
        :return[EquineOutput] : Output object containing prediction probabilities and OOD scores
        """
        X_embed = self.model.compute_embeddings(X)
        if X_embed.shape == torch.Size([self.model.emb_out_dim]):
            X_embed = X_embed.unsqueeze(dim=0)  # Handle single examples
        preds, dists = self.model(X)
        if self.use_temperature:
            dists = dists / self.temperature
            preds = torch.softmax(torch.negative(dists), dim=1)
        ood_dist = self._compute_ood_dist(X_embed, preds, dists)
        ood_scores = self._compute_outlier_scores(ood_dist, preds)

        return EquineOutput(classes=preds, ood_scores=ood_scores, embeddings=X_embed)

    @icontract.require(lambda calib_frac: (calib_frac > 0.0) and (calib_frac < 1.0))
    def update_support(
        self, support_x: torch.Tensor, support_y: torch.Tensor, calib_frac: float
    ) -> None:
        """Function to update protonet support examples with given examples
        :param support_x: Tensor containing support examples for protonet
        :param support_y: Tensor containing labels for given support examples
        :param calib_frac: Fraction of given support data to use for OOD calibration
        """

        support_x, calib_x, support_y, calib_y = train_test_split(
            support_x, support_y, test_size=calib_frac, stratify=support_y
        )
        labels, counts = torch.unique(support_y, return_counts=True)
        support = OrderedDict()
        for label, count in list(zip(labels.tolist(), counts.tolist())):
            class_support = generate_support(
                support_x,
                support_y,
                support_size=count,
                selected_labels=[label],
            )
            support.update(class_support)

        self.model.update_support(support)

        X_embed = self.model.compute_embeddings(calib_x)
        preds, dists = self.model(calib_x)
        ood_dists = self._compute_ood_dist(X_embed, preds, dists)

        self._fit_outlier_scores(ood_dists, calib_y)

    def save(self, path: str) -> None:
        """Function to save all model parameters to a file
        :param path: Filename to write the model
        """
        model_settings = {
            "cov_type": self.cov_type,
            "emb_out_dim": self.emb_out_dim,
            "use_temperature": self.use_temperature,
            "init_temperature": self.temperature.item(),
        }

        jit_model = torch.jit.script(self.model.embedding_model)
        buffer = io.BytesIO()
        torch.jit.save(jit_model, buffer)
        buffer.seek(0)

        save_data = {
            "settings": model_settings,
            "support": self.model.support,
            "outlier_kde": self.outlier_score_kde,
            "model_head_save": self.model.model_head.state_dict(),
            "embed_jit_save": buffer,
            "train_summary": self.train_summary,
        }

        torch.save(save_data, path)  # TODO allow model checkpointing

    @classmethod
    def load(cls, path: str) -> Equine:  # noqa: F821 TODO typehint doesnt want to work?
        """Function to load previously saved EquineProtonet model
        :param path: input filename
        :return[EquineGP] : The reconsituted EquineProtonet object
        """
        model_save = torch.load(path)
        support = model_save["support"]
        jit_model = torch.jit.load(model_save["embed_jit_save"])
        eq_model = cls(jit_model, **model_save["settings"])

        eq_model.model.model_head.load_state_dict(model_save["model_head_save"])
        eq_model.eval()
        eq_model.model.update_support(support)
        eq_model.outlier_score_kde = model_save["outlier_kde"]
        eq_model.train_summary = model_save["train_summary"]

        return eq_model



# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import equine as eq
import torch
from hypothesis import given
from hypothesis import strategies as st
import hypothesis.extra.numpy as hnp


@st.composite
def support_dataset(draw):
    support_sz = draw(st.integers(min_value=2, max_value=25))
    task_count = draw(st.integers(min_value=2, max_value=10))
    tasks = [x for x in range(task_count)]
    dataset_row_count = draw(
        st.integers(min_value=(support_sz + 1) * task_count, max_value=1000)
    )
    dataset_col_count = draw(st.integers(min_value=1, max_value=1000))
    shape = (dataset_row_count, dataset_col_count)
    dataset_x = torch.rand(shape)
    dataset_y = []
    for task in tasks:
        dataset_y += [task] * (support_sz + 1)

    if len(dataset_x) > len(dataset_y):
        pad = (
            torch.randint(0, task_count, (len(dataset_x) - len(dataset_y), 1))
            .flatten()
            .tolist()
        )
        dataset_y += pad

    dataset_y = torch.Tensor(dataset_y)
    way = draw(st.integers(min_value=2, max_value=task_count))

    return dataset_x, dataset_y, support_sz, tasks, way


@given(dataset=support_dataset())
def test_generate_support(dataset) -> None:
    train_x, train_y, support_sz, tasks, way = dataset
    eq.utils.generate_support(train_x, train_y, support_sz, tasks)


@st.composite
def draw_two_tensors(draw):
    num_classes = draw(st.integers(min_value=2, max_value=128))
    num_examples = draw(st.integers(min_value=2, max_value=100))
    yh = draw(
        hnp.arrays(
            shape=st.just((num_examples, num_classes)),
            dtype="float32",
            elements=st.floats(
                allow_nan=False,
                allow_infinity=False,
                width=32,
                min_value=9.9999998245167e-15,
                max_value=1.0,
            ),
        ).map(torch.tensor)
    )
    yh = yh / yh.sum(dim=1).unsqueeze(-1)
    true_class = draw(
        st.lists(
            st.integers(min_value=0, max_value=num_classes - 1),
            min_size=num_examples,
            max_size=num_examples,
        ).map(torch.tensor)
    )
    return yh, true_class


@given(draw_two_tensors())
def test_brier_score(two_tensors) -> None:
    yh, yt = two_tensors
    assert eq.utils.brier_score(yh, yt) >= 0.0


@given(draw_two_tensors())
def test_brier_skill_score(two_tensors) -> None:
    yh, yt = two_tensors
    assert eq.utils.brier_skill_score(yh, yt) <= 1.0


@given(draw_two_tensors())
def test_ece(two_tensors) -> None:
    yh, yt = two_tensors
    assert eq.utils.expected_calibration_error(yh, yt) <= 1.0


@given(draw_two_tensors())
def test_metric_summary(two_tensors) -> None:
    yh, yt = two_tensors
    eq_out = eq.EquineOutput(yh, torch.ones(yt.size()), torch.ones(yt.size()))
    metrics = eq.utils.generate_model_metrics(eq_out, yt)
    assert "accuracy" in metrics
    assert "microF1Score" in metrics
    assert "brierScore" in metrics
    assert "brierSkillScore" in metrics
    assert "expectedCalibrationError" in metrics



# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from __future__ import annotations
import torch
from torch.utils.data import TensorDataset  # type: ignore
from abc import ABC, abstractmethod
from typing import Any

from .equine_output import EquineOutput


class Equine(torch.nn.Module, ABC):
    """EQUINE Abstract Base Class (ABC):
    EQUINE is set up to extend torch's nn.Module to enrich it with
    a method that enables uncertainty quantification and visualization. Most
    importantly, the `.predict()` method must be outfitted to return an
    EquineOutput object that contains both the class logits
    *and* an out-of-distribution (ood) score.
    """

    def __init__(self, embedding_model, head_layers=1) -> None:
        super().__init__()
        self.embedding_model = embedding_model
        self.head_layers = head_layers

    @abstractmethod
    def forward(self, X: torch.Tensor):
        raise NotImplementedError

    @abstractmethod
    def predict(self, X: torch.Tensor) -> EquineOutput:
        raise NotImplementedError

    def train_model(self, dataset: TensorDataset, **kwargs) -> dict[str, Any]:
        raise NotImplementedError

    def save(self, path: str) -> None:
        raise NotImplementedError

    @classmethod
    def load(cls: Equine, path: str) -> Equine:  # noqa: F821 # type: ignore
        raise NotImplementedError



# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import math
import io

import icontract
import torch
from torch.utils.data import TensorDataset, DataLoader  # type: ignore
from typing import Optional, Callable
from tqdm import tqdm
from typeguard import typechecked
from datetime import datetime
from sklearn.model_selection import train_test_split

from .equine import Equine, EquineOutput
from .utils import generate_train_summary


# -------------------------------------------------------------------------------
# Note that the below code for
# * `_random_ortho`,
# * `_RandomFourierFeatures``, and
# * `_Laplace`
# is copied and modified from https://github.com/y0ast/DUE/blob/main/due/sngp.py
# under its original MIT license, redisplayed here:
# -------------------------------------------------------------------------------
# MIT License
#
# Copyright (c) 2021 Joost van Amersfoort
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
# ------------------------------------------------------------------------------
# Following the recommendation of their README at https://github.com/y0ast/DUE
# we encourage anyone using this code in their research to cite the following papers:
#
# @article{van2021on,
#   title={On Feature Collapse and Deep Kernel Learning for Single Forward Pass Uncertainty},
#   author={van Amersfoort, Joost and Smith, Lewis and Jesson, Andrew and Key, Oscar and Gal, Yarin},
#   journal={arXiv preprint arXiv:2102.11409},
#   year={2021}
# }
#
# @article{liu2020simple,
#  title={Simple and principled uncertainty estimation with deterministic deep learning via distance awareness},
#  author={Liu, Jeremiah and Lin, Zi and Padhy, Shreyas and Tran, Dustin and Bedrax Weiss, Tania and Lakshminarayanan, Balaji},
#  journal={Advances in Neural Information Processing Systems},
#  volume={33},
#  pages={7498--7512},
#  year={2020}
# }


@typechecked
def _random_ortho(n: int, m: int) -> torch.Tensor:
    q, _ = torch.linalg.qr(torch.randn(n, m))
    return q


@typechecked
class _RandomFourierFeatures(torch.nn.Module):
    def __init__(
        self, in_dim: int, num_random_features: int, feature_scale: Optional[float]
    ) -> None:
        super().__init__()
        if feature_scale is None:
            feature_scale = math.sqrt(num_random_features / 2)

        self.register_buffer("feature_scale", torch.tensor(feature_scale))

        if num_random_features <= in_dim:
            W = _random_ortho(in_dim, num_random_features)
        else:
            # generate blocks of orthonormal rows which are not neccesarily orthonormal
            # to each other.
            dim_left = num_random_features
            ws = []
            while dim_left > in_dim:
                ws.append(_random_ortho(in_dim, in_dim))
                dim_left -= in_dim
            ws.append(_random_ortho(in_dim, dim_left))
            W = torch.cat(ws, 1)

        feature_norm = torch.randn(W.shape) ** 2
        W = W * feature_norm.sum(0).sqrt()
        self.register_buffer("W", W)

        b = torch.empty(num_random_features).uniform_(0, 2 * math.pi)
        self.register_buffer("b", b)

    def forward(self, x) -> torch.Tensor:
        k = torch.cos(x @ self.W + self.b)
        k = k / self.feature_scale

        return k


class _Laplace(torch.nn.Module):
    def __init__(
        self,
        feature_extractor: torch.nn.Module,
        num_deep_features: int,
        num_gp_features: int,
        normalize_gp_features: bool,
        num_random_features: int,
        num_outputs: int,
        feature_scale: Optional[float],
        mean_field_factor: Optional[float],  # required for classification problems
        ridge_penalty: float = 1.0,
    ) -> None:
        super().__init__()
        self.feature_extractor = feature_extractor
        self.mean_field_factor = mean_field_factor
        self.ridge_penalty = ridge_penalty
        self.train_batch_size = 0  # to be set later

        if num_gp_features > 0:
            self.num_gp_features = num_gp_features
            self.register_buffer(
                "random_matrix",
                torch.normal(0, 0.05, (num_gp_features, num_deep_features)),
            )
            self.jl = lambda x: torch.nn.functional.linear(x, self.random_matrix)  # type: ignore
        else:
            self.num_gp_features = num_deep_features
            self.jl = torch.nn.Identity()

        self.normalize_gp_features = normalize_gp_features
        if normalize_gp_features:
            self.normalize = torch.nn.LayerNorm(num_gp_features)

        self.rff = _RandomFourierFeatures(
            num_gp_features, num_random_features, feature_scale
        )
        self.beta = torch.nn.Linear(num_random_features, num_outputs)

        self.num_data = 0  # to be set later
        self.register_buffer("seen_data", torch.tensor(0))

        precision = torch.eye(num_random_features) * self.ridge_penalty
        self.register_buffer("precision", precision)

        self.recompute_covariance = True
        self.register_buffer("covariance", torch.eye(num_random_features))
        self.training_parameters_set = False

    def reset_precision_matrix(self):
        identity = torch.eye(self.precision.shape[0], device=self.precision.device)
        self.precision = identity * self.ridge_penalty
        self.seen_data = torch.tensor(0)
        self.recompute_covariance = True

    @icontract.require(lambda num_data: num_data > 0)
    @icontract.require(
        lambda num_data, batch_size: (0 < batch_size) & (batch_size <= num_data)
    )
    def set_training_params(self, num_data, batch_size) -> None:
        self.num_data = num_data
        self.train_batch_size = batch_size
        self.training_parameters_set = True

    @icontract.require(lambda self: self.mean_field_factor is not None)
    def mean_field_logits(self, logits, pred_cov):
        # Mean-Field approximation as alternative to MC integration of Gaussian-Softmax
        # Based on: https://arxiv.org/abs/2006.07584

        logits_scale = torch.sqrt(1.0 + torch.diag(pred_cov) * self.mean_field_factor)
        if self.mean_field_factor > 0:  # type: ignore
            logits = logits / logits_scale.unsqueeze(-1)

        return logits

    @icontract.require(lambda self: self.training_parameters_set)
    def forward(self, x):
        f = self.feature_extractor(x)
        f_reduc = self.jl(f)
        if self.normalize_gp_features:
            f_reduc = self.normalize(f_reduc)

        k = self.rff(f_reduc)

        pred = self.beta(k)

        if self.training:
            precision_minibatch = k.t() @ k
            self.precision += precision_minibatch
            self.seen_data += x.shape[0]

            assert (
                self.seen_data <= self.num_data
            ), "Did not reset precision matrix at start of epoch"
        else:
            assert self.seen_data > (
                self.num_data - self.train_batch_size
            ), "Not seen sufficient data for precision matrix"

            if self.recompute_covariance:
                with torch.no_grad():
                    eps = 1e-7
                    jitter = eps * torch.eye(
                        self.precision.shape[1],
                        device=self.precision.device,
                    )
                    u, info = torch.linalg.cholesky_ex(self.precision + jitter)
                    assert (info == 0).all(), "Precision matrix inversion failed!"
                    torch.cholesky_inverse(u, out=self.covariance)  # type: ignore

                self.recompute_covariance = False

            with torch.no_grad():
                pred_cov = k @ ((self.covariance @ k.t()) * self.ridge_penalty)

            if self.mean_field_factor is None:
                return pred, pred_cov
            else:
                pred = self.mean_field_logits(pred, pred_cov)

        return pred


# -------------------------------------------------------------------------------
# EquineGP, below, demonstrates how to adapt that approach in EQUINE
@typechecked
class EquineGP(Equine):
    """
    An example of an EQUINE model that builds upon the approach in "Spectral Norm
    Gaussian Processes". This wraps any pytorch embedding neural network and provides
    the `forward`, `predict`, `save`, and `load` methods required by Equine.
    """

    def __init__(
        self,
        embedding_model: torch.nn.Module,
        emb_out_dim: int,
        num_classes: int,
        use_temperature: bool = False,
        init_temperature: float = 1.0,
        device: str = "cpu",
    ) -> None:
        """EquineGP constructor
        :param embedding_model: Neural Network feature embedding
        :param emb_out_dim: The number of deep features from the feature embedding
        :param num_classes: The number of output classes this model predicts
        :param use_temperature: whether to use temperature scaling after training
        :param init_temperature: what to use as the initial temperature (1.0 has no effect)
        "param device: either 'cuda' or 'cpu'
        """
        super().__init__(embedding_model)
        self.num_deep_features = emb_out_dim
        self.num_gp_features = emb_out_dim
        self.normalize_gp_features = True
        self.num_random_features = 1024
        self.num_outputs = num_classes
        self.mean_field_factor = 25
        self.ridge_penalty = 1
        self.feature_scale = 2
        self.use_temperature = use_temperature
        self.init_temperature = init_temperature
        self.register_buffer(
            "temperature", torch.Tensor(self.init_temperature * torch.ones(1))
        )
        self.model = _Laplace(
            self.embedding_model,
            self.num_deep_features,
            self.num_gp_features,
            self.normalize_gp_features,
            self.num_random_features,
            self.num_outputs,
            self.feature_scale,
            self.mean_field_factor,
            self.ridge_penalty,
        )
        self.device_type = device
        self.device = torch.device(self.device_type)

    def train_model(
        self,
        dataset: TensorDataset,
        loss_fn: Callable,
        opt: torch.optim.Optimizer,
        num_epochs: int,
        batch_size: int = 64,
        calib_frac: float = 0.1,
        num_calibration_epochs: int = 2,
        calibration_lr: float = 0.01,
    ):
        """Train or fine-tune an EquineGP model
        :param dataset: An iterable, pytorch TensorDataset
        :param loss_fn: A pytorch loss function, eg., torch.nn.CrossEntropyLoss()
        :param opt: A pytorch optimizer, e.g., torch.optim.Adam()
        :param num_epochs: The desired number of epochs to use for training,
        :param batch_size: The number of samples to use per batch
        :param calib_frac: fraction of training data to use in temperature scaling
        :param num_calibration_epochs: The desired number of epochs to use for temperature scaling,
        :param calibration_lr: learning rate for temperature scaling
        :return: A tuple containing the training history and a dataloader for the calibration data
        """

        if self.use_temperature:
            X, Y = dataset[:]
            train_x, calib_x, train_y, calib_y = train_test_split(
                X, Y, test_size=calib_frac, stratify=Y
            )  # TODO: Replace sklearn with torch call
            dataset = TensorDataset(train_x, train_y)
            self.temperature = torch.Tensor(
                self.init_temperature * torch.ones(1)
            ).type_as(self.temperature)

        train_loader = DataLoader(
            dataset, batch_size=batch_size, shuffle=True, drop_last=True
        )
        self.model.set_training_params(
            len(train_loader.sampler), train_loader.batch_size
        )
        self.model.train()
        for _ in tqdm(range(num_epochs)):
            self.model.reset_precision_matrix()
            epoch_loss = 0.0
            for i, (xs, labels) in enumerate(train_loader):
                opt.zero_grad()
                xs = xs.to(self.device)
                labels = labels.to(self.device)
                yhats = self.model(xs)
                loss = loss_fn(yhats, labels.to(torch.long))
                loss.backward()
                opt.step()
                epoch_loss += loss.item()
        self.model.eval()

        calibration_loader = None
        if self.use_temperature:
            dataset_calibration = TensorDataset(calib_x, calib_y)
            calibration_loader = DataLoader(
                dataset_calibration,
                batch_size=batch_size,
                shuffle=True,
                drop_last=False,
            )
            self.calibrate_temperature(
                calibration_loader, num_calibration_epochs, calibration_lr
            )

        _, train_y = dataset[:]
        date_trained = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        self.train_summary = generate_train_summary(self, train_y, date_trained)

        return self.train_summary, calibration_loader

    def calibrate_temperature(
        self,
        calibration_loader: DataLoader,
        num_calibration_epochs: int = 1,
        calibration_lr: float = 0.01,
    ) -> None:
        """
        Fine-tune the temperature after training.  Note this function is also run at the conclusion of train_model
        :param calibration_loader: data loader returned by train_model
        :param num_calibration_epochs: number of epochs to tune temperature
        :param calibration_lr: learning rate for temperature optimization
        :return:
        """
        self.temperature.requires_grad = True
        loss_fn = torch.nn.functional.cross_entropy
        optimizer = torch.optim.Adam([self.temperature], lr=calibration_lr)
        for _ in range(num_calibration_epochs):
            for (xs, labels) in calibration_loader:
                optimizer.zero_grad()
                xs = xs.to(self.device)
                labels = labels.to(self.device)
                with torch.no_grad():
                    logits = self.model(xs)
                logits = logits / self.temperature
                loss = loss_fn(logits, labels.to(torch.long))
                loss.backward()
                optimizer.step()
        self.temperature.requires_grad = False

    def forward(self, X: torch.Tensor) -> torch.Tensor:
        """EquineGP forward function, generates logits for classification
        :param X: Input tensor for generating predictions
        :return[torch.Tensor]: Output probabilities computed
        """
        X = X.to(self.device)
        with torch.no_grad():
            preds = self.model(X)
        return preds / self.temperature

    @icontract.ensure(
        lambda result: all((0 <= result.ood_scores) & (result.ood_scores <= 1.0))
    )
    def predict(self, X: torch.Tensor) -> EquineOutput:
        """Predict function for EquineGP, inherited and implemented from Equine
        :param X: Input tensor
        :return[EquineOutput] : Output object containing prediction probabilities and OOD scores
        """
        X = X.to(self.device)
        with torch.no_grad():
            logits = self.model(X)
        logits = logits / self.temperature
        preds = torch.softmax(logits, dim=1)
        equiprobable = torch.ones(self.num_outputs) / self.num_outputs
        max_entropy = torch.sum(torch.special.entr(equiprobable))
        ood_score = torch.sum(torch.special.entr(preds), dim=1) / max_entropy
        eq_out = EquineOutput(
            classes=preds, ood_scores=ood_score, embeddings=torch.Tensor([])
        )  # TODO return embeddings
        return eq_out

    def save(self, path: str) -> None:
        """Function to save all model parameters to a file
        :param path: Filename to write the model
        """
        model_settings = {
            "emb_out_dim": self.num_deep_features,
            "num_classes": self.num_outputs,
            "use_temperature": self.use_temperature,
            "init_temperature": self.temperature.item(),
            "device": self.device_type,
        }

        jit_model = torch.jit.script(self.model.feature_extractor)  # type: ignore
        buffer = io.BytesIO()
        torch.jit.save(jit_model, buffer)  # type: ignore
        buffer.seek(0)

        laplace_sd = self.model.state_dict()
        keys_to_delete = []
        for key in laplace_sd:
            if "feature_extractor" in key:
                keys_to_delete.append(key)
        for key in keys_to_delete:
            del laplace_sd[key]

        save_data = {
            "settings": model_settings,
            "num_data": self.model.num_data,
            "train_batch_size": self.model.train_batch_size,
            "laplace_model_save": laplace_sd,
            "embed_jit_save": buffer,
            "train_summary": self.train_summary,
        }

        torch.save(save_data, path)  # TODO allow model checkpointing

    @classmethod
    def load(cls, path: str) -> Equine:  # noqa: F821 # type: ignore
        """Function to load previously saved EquineGP model
        :param path: input filename
        :return[EquineGP] : The reconsituted EquineGP object
        """
        model_save = torch.load(path)
        jit_model = torch.jit.load(model_save["embed_jit_save"])  # type: ignore
        eq_model = cls(jit_model, **model_save["settings"])

        eq_model.train_summary = model_save["train_summary"]
        eq_model.model.load_state_dict(model_save["laplace_model_save"], strict=False)
        eq_model.model.seen_data = model_save["laplace_model_save"]["seen_data"]

        eq_model.model.set_training_params(
            model_save["num_data"], model_save["train_batch_size"]
        )
        eq_model.eval()

        return eq_model



# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
import torch
from hypothesis import given, settings, strategies as st
import pytest
import os

import equine as eq
from conftest import BasicEmbeddingModel, random_dataset


@given(
    data_shape=st.tuples(
        st.integers(min_value=1, max_value=1000),
        st.integers(min_value=1, max_value=1000),
    ),
    num_classes=st.integers(min_value=2, max_value=256),
)
def test_compute_embeddings(data_shape, num_classes):
    queries = torch.rand(data_shape)
    embed_model = BasicEmbeddingModel(data_shape[1], num_classes)
    model = eq.EquineProtonet(embed_model, num_classes)
    model.model.compute_embeddings(queries)


@st.composite
def embeddings(draw):
    dataset_row_count = draw(st.integers(min_value=100, max_value=1000))
    dataset_col_count = draw(st.integers(min_value=3, max_value=10))
    shape = (dataset_row_count, dataset_col_count)
    support_embeddings = torch.rand(shape)
    query_embeddings = torch.rand(shape)
    ptrs = [0]
    curr_idx = 0

    for i in range(dataset_col_count - 1):
        curr_idx += int(dataset_row_count / dataset_col_count)
        ptrs.append(curr_idx)

    ptrs.append(dataset_row_count - 1)
    ptrs = torch.Tensor(ptrs).to(torch.long)

    return support_embeddings, ptrs, query_embeddings


@given(random_dataset=random_dataset())
@settings(deadline=None)
def test_train_episodes(random_dataset):
    dataset, num_classes, way = random_dataset
    num_shot = 3
    num_episodes = 10
    episode_size = 512

    X, Y = dataset.tensors
    num_deep_features = 32
    embed_model = BasicEmbeddingModel(X.shape[1], num_deep_features)
    model = eq.EquineProtonet(embed_model, num_deep_features)
    model.train_model(
        dataset,
        way=way,
        support_size=num_shot,
        num_episodes=num_episodes,
        episode_size=episode_size,
    )

    assert model.model.training is False, "Model leaves training mode"
    assert len(model.model.support) == num_classes  # type: ignore
    # Test on multiple predictions
    eq_out = model.predict(X)
    assert len(eq_out.classes) == len(X)
    assert len(eq_out.ood_scores) == len(X)
    # Test on single prediction
    pred_out = model(X[0])
    assert len(pred_out) == 1, "Single prediction works"
    eq_out = model.predict(X[0])
    assert len(eq_out.classes) == 1, "Single prediction works"


@given(random_dataset=random_dataset())
@settings(deadline=None)
def test_train_episodes_with_temperature(random_dataset):
    dataset, num_classes, way = random_dataset
    num_shot = 3
    num_episodes = 10
    episode_size = 512

    X, Y = dataset.tensors
    num_deep_features = 32
    embed_model = BasicEmbeddingModel(X.shape[1], num_deep_features)
    model = eq.EquineProtonet(embed_model, num_deep_features, use_temperature=True)
    _, cal_x, cal_y = model.train_model(
        dataset,
        way=way,
        support_size=num_shot,
        num_episodes=num_episodes,
        episode_size=episode_size,
    )

    assert model.model.training is False, "Model leaves training mode"
    # Test on multiple predictions
    eq_out = model.predict(X)
    assert len(eq_out.classes) == len(X)
    assert len(eq_out.ood_scores) == len(X)
    # Test on single prediction
    pred_out = model(X[0])
    assert len(pred_out) == 1, "Single prediction works"
    eq_out = model.predict(X[0])
    assert len(eq_out.classes) == 1, "Single prediction works"

    model.calibrate_temperature(cal_x, cal_y, 1, 0.01)


@given(random_dataset=random_dataset())
def test_predict_fail_before_training(random_dataset):
    dataset, num_classes, _ = random_dataset
    X, _ = dataset.tensors
    embed_model = BasicEmbeddingModel(X.shape[1], num_classes)
    model = eq.EquineProtonet(embed_model, num_classes)
    with pytest.raises(ValueError):
        model(X)
    with pytest.raises(ValueError):
        model.predict(X)


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=2)
def test_equine_protonet_save_load(random_dataset) -> None:
    dataset, num_classes, _ = random_dataset
    X, Y = dataset.tensors
    embedding_model = BasicEmbeddingModel(X.shape[1], num_classes)

    model = eq.EquineProtonet(embedding_model, num_classes)
    model.train_model(dataset, num_episodes=10)

    old_output = model.predict(X[1:10])
    tmp_filename = "tmp_eq_proto.pt"
    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)
    model.save(tmp_filename)
    new_model = eq.EquineProtonet.load(tmp_filename)
    new_output = new_model.predict(X[1:10])
    assert (
        torch.nn.functional.mse_loss(old_output.classes, new_output.classes) <= 1e-7
    ), "Predictions changed on reload"
    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=2)
def test_equine_protonet_save_load_with_temperature(random_dataset) -> None:
    dataset, num_classes, _ = random_dataset
    X, Y = dataset.tensors
    embedding_model = BasicEmbeddingModel(X.shape[1], num_classes)

    model = eq.EquineProtonet(embedding_model, num_classes, use_temperature=True)
    model.train_model(dataset, num_episodes=10)

    old_output = model.predict(X[1:10])
    tmp_filename = "tmp_eq_proto.pt"
    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)
    model.save(tmp_filename)
    new_model = eq.EquineProtonet.load(tmp_filename)
    new_output = new_model.predict(X[1:10])
    assert (
        torch.nn.functional.mse_loss(old_output.classes, new_output.classes) <= 1e-7
    ), "Predictions changed on reload"
    if os.path.exists(tmp_filename):
        os.remove(tmp_filename)  # Cleanup



# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any, List, Union, Tuple
import icontract
import torch
from typeguard import typechecked
from collections import OrderedDict
from sklearn.metrics import accuracy_score, f1_score, confusion_matrix
from torchmetrics.classification import MulticlassCalibrationError

from .equine_output import EquineOutput
from .equine import Equine


@typechecked
@icontract.require(lambda y_hat, y_test: y_hat.size(dim=0) == y_test.size(dim=0))
@icontract.ensure(lambda result: result >= 0.0)
def brier_score(y_hat: torch.Tensor, y_test: torch.Tensor) -> float:
    """
    Compute the Brier score for a multiclass problem
    :param y_hat:  Probabilities for each class
    :param y_test: Integer argument class labels (ground truth)
    :return[float]: Brier score
    """
    (_, num_classes) = y_hat.size()
    one_hot_y_test = torch.nn.functional.one_hot(y_test, num_classes=num_classes)
    bs = torch.mean(torch.sum((y_hat - one_hot_y_test) ** 2, dim=1)).item()
    return bs


@typechecked
@icontract.require(lambda y_hat, y_test: y_hat.size(dim=0) == y_test.size(dim=0))
@icontract.ensure(lambda result: result <= 1.0)
def brier_skill_score(y_hat: torch.Tensor, y_test: torch.Tensor) -> float:
    """Compute the Brier skill score as compared to randomly guessing
    :param y_hat: Probabilities for each class
    :param y_test: Integer argument class labels (ground truth)
    :return[float]: Brier skill score"""
    (_, num_classes) = y_hat.size()
    random_guess = (1.0 / num_classes) * torch.ones(y_hat.size())
    bs0 = brier_score(random_guess, y_test)
    bs1 = brier_score(y_hat, y_test)
    bss = 1.0 - bs1 / bs0
    return bss


@typechecked
@icontract.require(lambda y_hat, y_test: y_hat.size(dim=0) == y_test.size(dim=0))
@icontract.ensure(lambda result: (0.0 <= result) and (result <= 1.0))
def expected_calibration_error(y_hat: torch.Tensor, y_test: torch.Tensor) -> float:
    """Compute the expected calibration error (ECE) for a multiclass problem
    :param y_hat: Probabilities for each class
    :param y_test: class label indices (ground truth)
    :return[float]: Expected calibration error"""
    (_, num_classes) = y_hat.size()
    metric = MulticlassCalibrationError(num_classes=num_classes, n_bins=25, norm="l1")
    ece = metric(y_hat, y_test).item()
    return ece


@typechecked
@icontract.require(
    lambda train_y, selected_labels: len(selected_labels) <= len(train_y)
)
@icontract.ensure(
    lambda result, selected_labels: set(result.keys()).issubset(set(selected_labels))
)
def _get_shuffle_idxs_by_class(
    train_y: torch.Tensor, selected_labels: List
) -> dict[Any, torch.Tensor]:
    """
    Internal helper function to randomly select indices of example classes for a given
    set of labels
    :param train_y: Label data
    :param selected_labels: List of unique labels found in the label data
    :param train_y: Label data
    :param selected_labels: List of unique labels found in the label data
    :return[dict[Any, torch.Tensor]]: Tensor of indices corresponding to each label
    """
    shuffled_idxs_by_class = OrderedDict()
    for label in selected_labels:
        label_idxs = torch.argwhere(train_y == label).squeeze()
        shuffled_idxs_by_class[label] = label_idxs[torch.randperm(label_idxs.shape[0])]

    return shuffled_idxs_by_class


@typechecked
@icontract.require(lambda train_x, train_y: len(train_x) <= len(train_y))
@icontract.require(
    lambda selected_labels, train_x: (0 < len(selected_labels))
    & (len(selected_labels) < len(train_x))
)
@icontract.require(
    lambda support_size, train_x: (0 < support_size) & (support_size < len(train_x))
)
@icontract.require(
    lambda support_size, selected_labels, train_x: support_size * len(selected_labels)
    <= len(train_x)
)
@icontract.ensure(
    lambda result, selected_labels, return_indexes: len(result[0].keys())
    == len(selected_labels)
    if (return_indexes is True)
    else len(result.keys()) == len(selected_labels)
)
def generate_support(
    train_x: torch.Tensor,
    train_y: torch.Tensor,
    support_size: int,
    selected_labels: List,
    return_indexes=False,
) -> Union[
    dict[Any, torch.Tensor], Tuple[dict[Any, torch.Tensor], dict[Any, torch.Tensor]]
]:
    """Randomly select `support_size` examples of `way` classes from the examples in
    `train_x` with corresponding labels in `train_y` and return them as a dictionary
    :param train_x: Input training data
    :param train_y: Corresponding classification labels
    :param support_size:  Number of support examples for each class
    :param selected_labels: Selected class labels to generate examples from
    :param return_indexes: If True, also return the indices of the support examples
    :return[dict[Any, torch.Tensor]]: Ordered dictionary of class labels with corresponding support examples
    """
    labels, counts = torch.unique(train_y, return_counts=True)
    for label, count in list(zip(labels, counts)):
        if (label in selected_labels) and (count < support_size):
            raise ValueError(f"Not enough support examples in class {label}")

    shuffled_idxs = _get_shuffle_idxs_by_class(train_y, selected_labels)

    support = OrderedDict()
    for label in selected_labels:
        shuffled_x = train_x[shuffled_idxs[label]]

        assert torch.unique(train_y[shuffled_idxs[label]]).tolist() == [
            label
        ], "Not enough support for label " + str(label)
        selected_support = shuffled_x[:support_size]
        support[label] = selected_support

    if return_indexes:
        return support, shuffled_idxs
    else:
        return support


@typechecked
def generate_episode(
    train_x: torch.Tensor,
    train_y: torch.Tensor,
    support_size: int,
    way: int,
    episode_size: int,
) -> Tuple[dict[Any, torch.Tensor], torch.Tensor, torch.Tensor]:
    """
    Generate a single episode of data for a few-shot learning task
    :param train_x: Input training data
    :param train_y: Corresponding classification labels
    :param support_size:  Number of support examples for each class
    :param way: Number of classes in the episode
    :param episode_size: Total number of examples in the episode
    :return[Tuple[dict[Any, torch.Tensor], torch.Tensor, torch.Tensor]]: Tuple of support examples, query examples, and query labels
    """
    labels = torch.unique(train_y)
    selected_labels = sorted(
        labels[torch.randperm(labels.shape[0])][:way].tolist()
    )  # need to be in same order every time

    support, shuffled_idxs = generate_support(
        train_x, train_y, support_size, selected_labels, return_indexes=True
    )

    examples_per_task = int(episode_size / way)

    episode_data_list = []
    episode_label_list = []
    episode_support = OrderedDict()
    for episode_label, label in enumerate(selected_labels):
        shuffled_x = train_x[shuffled_idxs[label]]
        shuffled_y = torch.Tensor(
            [episode_label] * len(shuffled_idxs[label])
        )  # need sequential labels for episode

        num_remaining_examples = shuffled_x.shape[0] - support_size
        assert num_remaining_examples > 0, (
            "Cannot have "
            + str(num_remaining_examples)
            + " left with support_size "
            + str(support_size)
            + " and shape "
            + str(shuffled_x.shape)
            + " from train_x shaped "
            + str(train_x.shape)
        )
        episode_end_idx = support_size + min(num_remaining_examples, examples_per_task)

        episode_data_list.append(shuffled_x[support_size:episode_end_idx])
        episode_label_list.append(shuffled_y[support_size:episode_end_idx])
        episode_support[episode_label] = support[label]

    episode_x = torch.concat(episode_data_list)
    episode_y = torch.concat(episode_label_list)

    return episode_support, episode_x, episode_y.squeeze().to(torch.long)


@typechecked
@icontract.require(
    lambda eq_preds, true_y: eq_preds.classes.size(dim=0) == true_y.size(dim=0)
)
def generate_model_metrics(
    eq_preds: EquineOutput, true_y: torch.Tensor
) -> dict[str, Any]:
    pred_y = torch.argmax(eq_preds.classes, dim=1)
    metrics = {
        "accuracy": accuracy_score(true_y, pred_y),
        "microF1Score": f1_score(true_y, pred_y, average="micro"),
        "confusionMatrix": confusion_matrix(true_y, pred_y).tolist(),
        "brierScore": brier_score(eq_preds.classes, true_y),
        "brierSkillScore": brier_skill_score(eq_preds.classes, true_y),
        "expectedCalibrationError": expected_calibration_error(
            eq_preds.classes, true_y
        ),
    }
    return metrics


def get_num_examples_per_label(Y: torch.Tensor):
    tensor_labels, tensor_counts = Y.unique(return_counts=True)

    examples_per_label = []
    for i, label in enumerate(tensor_labels):
        examples_per_label.append(
            {"label": label.item(), "numExamples": tensor_counts[i].item()}
        )

    return examples_per_label


@icontract.require(lambda train_y: train_y.shape[0] > 0)
def generate_train_summary(model: Equine, train_y: torch.Tensor, date_trained: str):
    train_summary = {
        "numTrainExamples": get_num_examples_per_label(train_y),
        "dateTrained": date_trained,
        "modelType": model.__class__.__name__,
    }
    return train_summary


@icontract.require(
    lambda eq_preds, test_y: test_y.shape[0] == eq_preds.classes.shape[0]
)
def generate_model_summary(
    model: Equine,
    eq_preds: EquineOutput,
    test_y: torch.Tensor,
):
    summary = generate_model_metrics(eq_preds, test_y)
    summary["numTestExamples"] = get_num_examples_per_label(test_y)
    summary.update(model.train_summary)  # union of train_summary and generated metrics

    return summary



# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import equine as eq
import torch
from hypothesis import given, settings

import pytest
from conftest import BasicEmbeddingModel, random_dataset


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=5)
def test_equine_protonet_instantiation(random_dataset) -> None:
    dataset, num_classes, _ = random_dataset
    X, Y = dataset.tensors
    embedding_model = BasicEmbeddingModel(X.shape[1], num_classes)

    model = eq.EquineProtonet(embedding_model, num_classes)
    model.train_model(torch.utils.data.TensorDataset(X, Y), num_episodes=0)


@given(random_dataset=random_dataset())
@settings(deadline=None, max_examples=5)
def test_equine_gp_instantiation(random_dataset) -> None:
    dataset, num_classes, _ = random_dataset
    X, _ = dataset.tensors
    embedding_model = BasicEmbeddingModel(X.shape[1], num_classes)

    model = eq.EquineGP(embedding_model, num_classes, num_classes)
    loss_fn = torch.nn.CrossEntropyLoss()
    optimizer = torch.optim.SGD(
        model.parameters(),
        lr=0.001,
        momentum=0.9,
        weight_decay=0.0001,
    )
    model.train_model(dataset, loss_fn, optimizer, 10)


def test_unimplemented_errors():
    em = BasicEmbeddingModel(2, 4)
    with pytest.raises(TypeError):
        eq.Equine(em)
    with pytest.raises(NotImplementedError):
        eq.Equine.predict(None, torch.Tensor([1, 2]))
    with pytest.raises(NotImplementedError):
        eq.Equine.forward(None, torch.Tensor([1, 2]))
    with pytest.raises(NotImplementedError):
        eq.Equine.load("tmp")  # type: ignore



# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from torch import Tensor


@dataclass
class EquineOutput:
    classes: Tensor
    ood_scores: Tensor
    embeddings: Tensor
